# -*- coding: utf-8 -*-

"""
geodaisy.converters
~~~~~~~~~~~~~~~~

This module provides converters to and from GeoJSON, WKT, geo_interface, etc.

geodaisy.geo_object uses these converters; they can also be used directly.
"""

import json
import re
from ast import literal_eval
from typing import Union  # noqa: F401


geo_types = {'Point', 'MultiPoint', 'LineString', 'MultiLineString',
             'Polygon', 'MultiPolygon'}

wkt_types = {x.upper() for x in geo_types}

type_translations = {x.upper(): x for x in geo_types}


def geo_interface_to_wkt(geo_interface):
    # type: (dict) -> str
    """Converts a geo_interface dictionary to a Well Known Text string."""
    # Convert to string and change brackets to parentheses
    coords = str(geo_interface['coordinates'])
    coords = coords.replace('[', '(')
    coords = coords.replace(']', ')')

    # Remove commas within coordinate pairs
    coords = re.sub(r'(?<=\d),', '', coords)

    # Remove parentheses and commas separating coordinates
    coords = re.sub(r'(?<=\d)\), \(', ', ', coords)

    # Get the type and clean up extra parentheses in special cases
    geo_type = geo_interface['type'].upper()

    if geo_type == 'LINESTRING':
        coords = coords.replace('((', '(')
        coords = coords.replace('))', ')')
        coords = coords.replace('), (', ', ')
    elif geo_type in {'MULTILINESTRING', 'POLYGON'}:
        coords = coords.replace('(((', '((')
        coords = coords.replace(')))', '))')
        coords = coords.replace(')), ((', '), (')
    elif geo_type == 'MULTIPOINT':
        coords = coords.replace('((', '(')
        coords = coords.replace('))', ')')
    elif geo_type == 'MULTIPOLYGON':
        coords = coords.replace('((((', '(((')
        coords = coords.replace('))))', ')))')
        coords = coords.replace('))), (((', ')), ((')

    return '{} {}'.format(geo_type, coords)


def wkt_to_geo_interface(wkt):
    # type: (str) -> dict
    """Converts a WKT string to a geo_interface dictionary."""
    try:
        wkt_type, coords = re.split(r'(?<=[A-Z])\s', wkt)

        geo_type = type_translations[wkt_type]

        # Clean up the strings so they'll covert correctly
        if geo_type in {'Polygon', 'MultiLineString', 'MultiPolygon'}:
            coords = re.sub(r'(?<=\d)\), \((?=\d)', ')), ((', coords)

        # Pairs of coordinates must be enclosed in parentheses
        coords = re.sub(r'(?<=\d), (?=\d)', '), (', coords)

        # Coordinates within parentheses must be separated by commas
        coords = re.sub(r'(?<=\d) (?=\d)', ', ', coords)

        # Now we can turn the string into a tuple or a tuple of tuples
        coords = literal_eval(coords)

        coords = reformat_coordinates(coords, 'geo_interface')  # type: ignore  # noqa: E501

        # If we only have a simple polygon (no hole), the coordinate array
        # won't be deep enough to satisfy the GeoJSON/geo_interface spec, so
        # we need to enclose it in a list.
        numbers = {float, int}
        if geo_type == 'Polygon' and type(coords[0][0]) in numbers:
            coords = [coords]  # type: ignore
        elif geo_type == 'MultiPolygon' and type(coords[0][0][0]) in numbers:
            coords = [coords]  # type: ignore

    except Exception:
        raise ValueError('{} is not a WKT string'.format(wkt))

    return {'type': geo_type, 'coordinates': coords}


def dict_to_geo_interface(geo_dict):
    # type: (dict) -> dict
    """Converts a dictionary into a standardized geo_interface dictionary."""
    if not geo_dict.get('type') in geo_types:
        raise ValueError('A geo_interface-compatible dictionary must'
                         ' have a "type" key with one of the following'
                         'values: {}'.format(geo_types))
    try:
        coordinates = geo_dict['coordinates']
    except KeyError:
        raise KeyError('A geo_interface-compatible dictionary must'
                       ' have a "coordinates" key-value pair.')
    geo_dict['coordinates'] = reformat_coordinates(coordinates,
                                                   'geo_interface')

    return geo_dict


def geo_interface_to_geojson(geo_interface):
    # type: (dict) -> dict
    """Converts a geo_interface dictionary into a raw GeoJSON dictionary."""
    coords = reformat_coordinates(geo_interface['coordinates'], 'geojson')

    return {'type': geo_interface['type'], 'coordinates': coords}


def wkt_to_geojson(wkt):
    # type: (str) -> str
    """Converts a WKT string to serialized GeoJSON."""
    return json.dumps(geo_interface_to_geojson(wkt_to_geo_interface(wkt)))


def geojson_to_wkt(geojson):
    # type: (Union[str, dict]) -> str
    """Converts GeoJSON (serialize or raw) to a WKT string."""
    if isinstance(geojson, str):
        raw_geojson = json.loads(geojson)  # type: dict
    elif isinstance(geojson, dict):
        raw_geojson = geojson
    else:
        raise ValueError('{} is not GeoJSON'.format(geojson))

    return geo_interface_to_wkt(dict_to_geo_interface(raw_geojson))


def reformat_coordinates(item, style):
    # type: (Union[list, tuple], str) -> Union[list, tuple]
    """
    Converts tuples, tuples of tuples, lists of tuples, etc. into lists and
    lists of lists, etc. and preserves points/coordinate pairs as lists or
    tuples depending on the desired style.
    """
    if type(item) in {tuple, list} and type(item[0]) in {tuple, list}:
        return [reformat_coordinates(x, style) for x in item]
    else:
        if style == 'geojson':
            return list(item)
        elif style == 'geo_interface':
            return tuple(item)
        else:
            raise ValueError('style must be geojson or geo_interface.')



# -*- coding: utf-8 -*-

"""
Geodaisy is a Python library for creating geo objects represented by types
and coordinates and translating between various standards and
representations, including GeoJSON, Well Known Text, and Python's
__geo_interface__ protcol, which is used by other geo libraries.

Geodaisy is lightweight and does not rely on C/C++ dependencies like libgeos.

Usage example:
   >>> from geodaisy.geo_object import GeoObject
   >>> geo_obj = GeoObject(thing)
   >>> geo_obj.geojson
   {'coordinates': XXX, 'type': xxx}

For more information about the library and how to use it, see README.md.
:copyright: (c) 2018 by Kevin Brochet-Nguyen.
:license: MIT—see LICENSE for more details.
"""

from .geo_object import GeoObject  # noqa: F401



# -*- coding: utf-8 -*-

"""
geodaisy.geo_object
~~~~~~~~~~~~~~~~

This module provides a GeoObject object for creating Python geo objects that
that translate to and from other geo libraries, mappings, and specificaions.
"""

import json
from typing import Any  # noqa: F401

from .converters import (dict_to_geo_interface,
                         geo_interface_to_geojson,
                         geo_interface_to_wkt,
                         wkt_to_geo_interface,
                         wkt_types)


class GeoObject(object):
    """
    Creates an object from an object's geo_interface or from WKT or GeoJSON.

    GeoJSON (serialized and raw), WKT (Well Known Text), and normalized
    geo_interface representations of the object are available as attributes.
    """

    def __init__(self, geo_thing):
        # type: (Any) -> None
        """
        Creates a GeoObject if geo_thing has or represents a geo_interface.
        """
        geo_interface = self._get_geo_interface(geo_thing)
        self.type = geo_interface['type']
        self.coordinates = geo_interface['coordinates']

    def __repr__(self):
        # type: () -> str
        return str(self.__geo_interface__)

    __str__ = __repr__

    @property
    def __geo_interface__(self):
        # type: () -> dict
        """Returns the geo_interface representation of the object."""
        return {'type': self.type, 'coordinates': self.coordinates}

    def geojson(self):
        # type: () -> str
        """Returns the serialized GeoJSON representation of the object."""
        return json.dumps(self.rawgeojson())

    def rawgeojson(self):
        # type: () -> dict
        """Returns the unserialized GeoJSON representation of the object."""
        return geo_interface_to_geojson(self.__geo_interface__)

    def wkt(self):
        # type: () -> str
        """Returns the Well Known Text representation of the object."""
        return geo_interface_to_wkt(self.__geo_interface__)

    def _get_geo_interface(self, geo_thing):
        # type: (Any) -> dict
        """Attempts to get an object's geo_interface."""
        # Input might be GeoJSON or WKT
        if isinstance(geo_thing, str):
            geo_interface = self._parse_string(geo_thing)
        # Input might be raw GeoJSON or a geo_interface dictionary
        elif isinstance(geo_thing, dict):
            geo_interface = dict_to_geo_interface(geo_thing)
        # Input might be an object with a geo_interface
        else:
            try:
                geo_interface = dict_to_geo_interface(
                    geo_thing.__geo_interface__)
            except AttributeError:
                raise AttributeError('Object has no geo_interface.')

        return geo_interface

    def _parse_string(self, geo_thing):
        # type: (str) -> dict
        """Checks to see if the string is geojson or WKT."""
        error_msg = 'Strings must be valid GeoJSON or WKT'
        if geo_thing.startswith('{'):
            try:
                geo_dict = json.loads(geo_thing)
                return dict_to_geo_interface(geo_dict)
            except ValueError:
                raise ValueError(error_msg)
        else:
            wkt_type = geo_thing.split(' ')[0]
            if wkt_type not in wkt_types:
                raise ValueError(error_msg)
            else:
                return wkt_to_geo_interface(geo_thing)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'geodaisy'
DESCRIPTION = 'Python GeoJSON, WKT, and __geo_interface__ made easy.'
URL = 'https://github.com/kmbn/geodaisy'
EMAIL = 'kmbn@nevermindtheumlauts.com'
AUTHOR = 'Kevin Brochet-Nguyen'
REQUIRES_PYTHON = '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*'
VERSION = '0.1.1'
REQUIRED = ['typing']  # Required packages

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Make sure 'README.md' is present in th MANIFEST.in!
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = [('test', None,
                     'Perform a test upload to test.pypi.org if set.')]

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        """Set default values for options."""
        # Each user option must be listed here with their default value.
        self.test = False

    def finalize_options(self):
        """Post-process options."""
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        if self.test:
            repository = '--repository-url https://test.pypi.org/legacy/ '
        else:
            repository = ''

        self.status('Building Source distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        files = os.listdir('dist/')
        source_dist = [f for f in files if f.endswith('tar.gz')][0]
        wheel_dist = [f for f in files if f.endswith('whl')][0]

        self.status('Uploading Source distribution to PyPi via Twine…')
        os.system('twine upload {0}dist/{1}'.format(repository, source_dist))

        self.status('Uploading Wheel distribution to PyPi via Twine…')
        os.system('twine upload {0}dist/{1}'.format(repository, wheel_dist))

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',  # This is important!
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=REQUIRED,
    setup_requires=['setuptools>=38.6.0'],
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering :: GIS',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)
